//------------------------------------------------------------------------------------------------
//
//   SG Craft - Stargate base tile entity
//
//------------------------------------------------------------------------------------------------

package gcewing.sg;

import java.util.*;
import java.lang.reflect.Method;
import org.apache.logging.log4j.*;
import io.netty.channel.*;

import net.minecraft.entity.*;
import net.minecraft.entity.player.*;
import net.minecraft.inventory.*;
import net.minecraft.item.*;
import net.minecraft.nbt.*;
import net.minecraft.network.*;
import net.minecraft.network.play.server.*;
import net.minecraft.potion.*;
import net.minecraft.server.*;
import net.minecraft.server.management.*;
import net.minecraft.tileentity.*;
import net.minecraft.util.*;
import net.minecraft.world.*;
import net.minecraft.world.chunk.*;

import net.minecraftforge.common.*;
import net.minecraftforge.common.util.*;
import net.minecraftforge.common.network.*;
import cpw.mods.fml.common.registry.*;
import cpw.mods.fml.common.network.*;
import cpw.mods.fml.relauncher.*;

import static gcewing.sg.BaseUtils.*;
import static gcewing.sg.Utils.*;

public class SGBaseTE extends BaseChunkLoadingTE  /*implements IInventory*/ {

	static boolean debugState = false;
	static boolean debugEnergyUse = false;
	static boolean debugConnect = false;

	public final static String symbolChars = SGAddressing.symbolChars;
	public final static int numRingSymbols = SGAddressing.numSymbols;
	public final static double ringSymbolAngle = 360.0 / numRingSymbols;
	
	final static int[] diallingTime = {40, 28}; // ticks
	final static int[] interDiallingTime = {10, 11}; // ticks
	final static String[] diallingSound = {"gcewing_sg:sg_dial7", "gcewing_sg:sg_dial9"};
	final static int transientDuration = 20; // ticks
	final static int disconnectTime = 30; // ticks
	
	final static double openingTransientIntensity = 1.3;
	final static double openingTransientRandomness = 0.25;
	final static double closingTransientRandomness = 0.25;
	final static double transientDamageRate = 50;
	
	// Configuration options
	static double maxEnergyBuffer = 1000;
	static double energyPerFuelItem = 96000;
	static double distanceFactorMultiplier = 1.0;
	static double interDimensionMultiplier = 4.0;
	static int gateOpeningsPerFuelItem = 24;
	static int minutesOpenPerFuelItem = 80;
	static int secondsToStayOpen = 5 * 60;
	static boolean oneWayTravel = false;
	static boolean closeFromEitherEnd = true;
	static int chunkLoadingRange = 1;
	static boolean logStargateEvents = false;
	
	static double energyToOpen;
	static double energyUsePerTick;
	static int ticksToStayOpen;
	public static boolean transparency = true;
	
	static Random random = new Random();
	static DamageSource transientDamage = new TransientDamageSource();
	
	public boolean isMerged;
	public SGState state = SGState.Idle;
	public double ringAngle, lastRingAngle, targetRingAngle; // degrees
	public int firstEngagedChevron;
	public int numEngagedChevrons;
	public String dialledAddress = "";
	//public String dialledAddress =  "MYNCRFT"; // "AAAAAAA";
	public boolean isLinkedToController;
	public int linkedX, linkedY, linkedZ;
	public boolean hasChevronUpgrade;
	SGLocation connectedLocation;
	boolean isInitiator;
	int timeout;
	double energyInBuffer;
	double distanceFactor; // all energy use is multiplied by this
	
//	public static final int firstFuelSlot = 0;
//	public static final int numFuelSlots = 4;
//	public static final int firstUpgradeSlot = 4;
//	public static final int numUpgradeSlots = 0;
//	IInventory inventory = new InventoryBasic("Stargate", false, numFuelSlots + numUpgradeSlots);

	double ehGrid[][][];
	
	static Method onEntityRemoved = BaseReflectionUtils.getMethod(World.class, "onEntityRemoved",
		"func_72847_b", "(Lnet/minecraft/entity/Entity;)V", Entity.class);

	public static void configure(BaseConfiguration cfg) {
		energyPerFuelItem = cfg.getDouble("stargate", "energyPerFuelItem", energyPerFuelItem);
		gateOpeningsPerFuelItem = cfg.getInteger("stargate", "gateOpeningsPerFuelItem", gateOpeningsPerFuelItem);
		minutesOpenPerFuelItem = cfg.getInteger("stargate", "minutesOpenPerFuelItem", minutesOpenPerFuelItem);
		secondsToStayOpen = cfg.getInteger("stargate", "secondsToStayOpen", secondsToStayOpen);
		oneWayTravel = cfg.getBoolean("stargate", "oneWayTravel", oneWayTravel);
		closeFromEitherEnd = cfg.getBoolean("stargate", "closeFromEitherEnd", closeFromEitherEnd);
		//energyPerFuelItem = minutesOpenPerFuelItem * 60 * 20;
		//maxEnergyBuffer = 2 * energyPerFuelItem;
		maxEnergyBuffer = cfg.getDouble("stargate", "maxEnergyBuffer", maxEnergyBuffer);
		energyToOpen = energyPerFuelItem / gateOpeningsPerFuelItem;
		energyUsePerTick = energyPerFuelItem / (minutesOpenPerFuelItem * 60 * 20);
		distanceFactorMultiplier = cfg.getDouble("stargate", "distanceFactorMultiplier", distanceFactorMultiplier);
		interDimensionMultiplier = cfg.getDouble("stargate", "interDimensionMultiplier", interDimensionMultiplier);
		System.out.printf("SGBaseTE: energyPerFuelItem = %s\n", energyPerFuelItem);
		System.out.printf("SGBaseTE: energyToOpen = %s\n", energyToOpen);
		System.out.printf("SGBaseTE: energyUsePerTick = %s\n", energyUsePerTick);
		ticksToStayOpen = 20 * secondsToStayOpen;
		chunkLoadingRange = cfg.getInteger("options", "chunkLoadingRange", chunkLoadingRange);
		if (chunkLoadingRange < 0)
			chunkLoadingRange = 0;
		transparency = cfg.getBoolean("stargate", "transparency", transparency);
		logStargateEvents = cfg.getBoolean("options", "logStargateEvents", logStargateEvents);
	}

	@Override
	public AxisAlignedBB getRenderBoundingBox() {
		return AxisAlignedBB.getAABBPool().getAABB(
			xCoord - 2, yCoord, zCoord - 2, xCoord + 3, yCoord + 5, zCoord + 3);
	}

	@Override
	public boolean hasCustomInventoryName() {
		return false;
	}
	
	@Override
	public double getMaxRenderDistanceSquared() {
		return 32768.0;
	}

//	@Override
//	public boolean isItemValidForSlot(int slot, ItemStack stack) {
//		if (slot >= firstFuelSlot && slot < firstFuelSlot + numFuelSlots)
//			return stack.getItem() == SGCraft.naquadah;
//		if (slot >= firstUpgradeSlot && slot < firstUpgradeSlot + numUpgradeSlots)
//			return SGCraft.isValidStargateUpgrade(stack.getItem());
//		return true;
//	}

	@Override
	public void onAddedToWorld() {
		if (SGBaseBlock.debugMerge)
			System.out.printf("SGBaseTE.onAddedToWorld\n");
		updateChunkLoadingStatus();
	}
	
	void updateChunkLoadingStatus() {
		if (state != SGState.Idle) {
			int n = chunkLoadingRange;
			setForcedChunkRange(-n, -n, n, n);
		}
		else
			clearForcedChunkRange();
	}

	public static SGBaseTE at(IBlockAccess world, int x, int y, int z) {
		TileEntity te = world.getTileEntity(x, y, z);
		if (te instanceof SGBaseTE)
			return (SGBaseTE)te;
		else
			return null;
	}
	
	public static SGBaseTE at(SGLocation loc) {
		if (loc != null) {
			World world = /*DimensionManager.*/SGAddressing.getWorld(loc.dimension);
			if (world != null)
				return SGBaseTE.at(world, loc.x, loc.y, loc.z);
		}
		return null;
	}
	
	public static SGBaseTE at(IBlockAccess world, NBTTagCompound nbt) {
		return SGBaseTE.at(world, nbt.getInteger("x"), nbt.getInteger("y"), nbt.getInteger("z"));
	}
	
	void setMerged(boolean state) {
		if (isMerged != state) {
			isMerged = state;
			if (logStargateEvents) {
				String address = tryToGetHomeAddress();
				if (address != null) {
					Logger log = LogManager.getLogger();
					String action = isMerged ? "ADDED" : "REMOVED";
					String name = getWorldObj().getWorldInfo().getWorldName();
					log.info(String.format("STARGATE %s %s (%s,%s,%s) %s",
						action, name, xCoord, yCoord, zCoord, address));
				}
			}
		}
	}
	
	String tryToGetHomeAddress() {
		try {
			return getHomeAddress();
		}
		catch (SGAddressing.AddressingError e) {
			return null;
		}
	}

	public int dimension() {
		if (worldObj != null)
			return worldObj.provider.dimensionId;
		else
			return -999;
	}
	
	@Override
	public void readFromNBT(NBTTagCompound nbt) {
		super.readFromNBT(nbt);
		isMerged = nbt.getBoolean("isMerged");
		state = SGState.valueOf(nbt.getInteger("state"));
		targetRingAngle = nbt.getDouble("targetRingAngle");
		firstEngagedChevron = nbt.getInteger("firstEngagedChevron");
		numEngagedChevrons = nbt.getInteger("numEngagedChevrons");
		//homeAddress = nbt.getString("homeAddress");
		dialledAddress = nbt.getString("dialledAddress");
		isLinkedToController = nbt.getBoolean("isLinkedToController");
		linkedX = nbt.getInteger("linkedX");
		linkedY = nbt.getInteger("linkedY");
		linkedZ = nbt.getInteger("linkedZ");
		hasChevronUpgrade = nbt.getBoolean("hasChevronUpgrade");
		if (nbt.hasKey("connectedLocation"))
			connectedLocation = new SGLocation(nbt.getCompoundTag("connectedLocation"));
		isInitiator = nbt.getBoolean("isInitiator");
		timeout = nbt.getInteger("timeout");
		if (nbt.hasKey("energyInBuffer"))
			energyInBuffer = nbt.getDouble("energyInBuffer");
		else
			energyInBuffer = nbt.getInteger("fuelBuffer");
		distanceFactor = nbt.getDouble("distanceFactor");
	}

	@Override
	public void writeToNBT(NBTTagCompound nbt) {
		super.writeToNBT(nbt);
		nbt.setBoolean("isMerged", isMerged);
		nbt.setInteger("state", state.ordinal());
		nbt.setDouble("targetRingAngle", targetRingAngle);
		nbt.setInteger("firstEngagedChevron", firstEngagedChevron);
		nbt.setInteger("numEngagedChevrons", numEngagedChevrons);
		//nbt.setString("homeAddress", homeAddress);
		nbt.setString("dialledAddress", dialledAddress);
		nbt.setBoolean("isLinkedToController", isLinkedToController);
		nbt.setInteger("linkedX", linkedX);
		nbt.setInteger("linkedY", linkedY);
		nbt.setInteger("linkedZ", linkedZ);
		nbt.setBoolean("hasChevronUpgrade", hasChevronUpgrade);
		if (connectedLocation != null)
			nbt.setTag("connectedLocation", connectedLocation.toNBT());
		nbt.setBoolean("isInitiator", isInitiator);
		nbt.setInteger("timeout", timeout);
		nbt.setDouble("energyInBuffer", energyInBuffer);
		nbt.setDouble("distanceFactor", distanceFactor);
	}
	
	public NBTTagCompound nbtWithCoords() {
		NBTTagCompound nbt = new NBTTagCompound();
		nbt.setInteger("x", xCoord);
		nbt.setInteger("y", yCoord);
		nbt.setInteger("z", zCoord);
		return nbt;
	}
	
	public boolean isActive() {
		return state != SGState.Idle && state != SGState.Disconnecting;
	}
	
	static boolean isValidSymbolChar(String c) {
		return SGAddressing.isValidSymbolChar(c);
	}
	
	static char symbolToChar(int i) {
		return SGAddressing.symbolToChar(i);
	}
	
	static int charToSymbol(char c) {
		return SGAddressing.charToSymbol(c);
	}

	static int charToSymbol(String c) {
		return SGAddressing.charToSymbol(c);
	}
	
	public boolean applyChevronUpgrade(ItemStack stack, EntityPlayer player) {
		if (!getWorldObj().isRemote && !hasChevronUpgrade && stack.stackSize > 0) {
			System.out.printf("SGBaseTE.applyChevronUpgrade: Installing chevron upgrade\n");
			hasChevronUpgrade = true;
			stack.stackSize -= 1;
			markDirty();
			markBlockForUpdate();
		}
		return true;
	}
	
	int getNumChevrons() {
		//if (upgradePresent(SGCraft.sgChevronUpgrade))
		if (hasChevronUpgrade)
			return 9;
		else
			return 7;
	}

//	boolean upgradePresent(Item item) {
//		for (int i = firstUpgradeSlot; i < firstUpgradeSlot + numUpgradeSlots; i++)
//			if (getItemInSlot(i) == item)
//				return true;
//		return false;
//	}
	
	Item getItemInSlot(int slot) {
		ItemStack stack = getStackInSlot(slot);
		return stack != null ? stack.getItem() : null;
	}

	public String getHomeAddress() throws SGAddressing.AddressingError {
		return SGAddressing.addressForLocation(new SGLocation(this));
	}
	
	public SGBaseBlock getBlock() {
		return (SGBaseBlock)getBlockType();
	}
	
	public int getRotation() {
		//return getBlockMetadata() & SGBaseBlock.rotationMask;
		return getBlock().rotationInWorld(getBlockMetadata(), this);
	}
	
	public double interpolatedRingAngle(double t) {
		return Utils.interpolateAngle(lastRingAngle, ringAngle, t);
	}
	
	@Override
	public boolean canUpdate() {
		return true;
	}
	
	@Override
	public void updateEntity() {
		if (worldObj.isRemote)
			clientUpdate();
		else {
			serverUpdate();
			checkForEntitiesInPortal();
		}
	}
	
	String side() {
		return worldObj.isRemote ? "Client" : "Server";
	}
	
	void enterState(SGState newState, int newTimeout) {
		if (debugState)
			System.out.printf("SGBaseTE: at (%s, %s, %s) in dimension %s entering state %s with timeout %s\n", 
				xCoord, yCoord, zCoord, worldObj.provider.dimensionId, newState, newTimeout);
		SGState oldState = state;
		state = newState;
		timeout = newTimeout;
		markDirty();
		markBlockForUpdate();
		if ((oldState == SGState.Idle) != (newState == SGState.Idle)) {
			updateChunkLoadingStatus();
			worldObj.notifyBlocksOfNeighborChange(xCoord, yCoord, zCoord, getBlockType());
		}
	}
	
	public boolean isConnected() {
		return state == SGState.Transient || state == SGState.Connected || state == SGState.Disconnecting;
	}
	
	DHDTE getLinkedControllerTE() {
		if (isLinkedToController) {
			TileEntity cte = worldObj.getTileEntity(linkedX, linkedY, linkedZ);
			if (cte instanceof DHDTE)
				return (DHDTE)cte;
		}
		return null;
	}
	
	void checkForLink() {
		int rangeXY = max(DHDTE.linkRangeX, DHDTE.linkRangeY);
		int rangeZ = DHDTE.linkRangeZ;
		if (SGBaseBlock.debugMerge)
			System.out.printf("SGBaseTE.checkForLink: in range +/-(%d,%d,%d) of (%d,%d,%d)\n",
				rangeXY, rangeZ, rangeXY, xCoord, yCoord, zCoord);
		for (int i = -rangeXY; i <= rangeXY; i++)
			for (int j = -rangeZ; j <= rangeZ; j++)
				for (int k = -rangeXY; k <= rangeXY; k++) {
					TileEntity te = worldObj.getTileEntity(xCoord + i, yCoord + j, zCoord + k);
					if (te instanceof DHDTE)
						((DHDTE)te).checkForLink();
				}
	}

	public void unlinkFromController() {
		if (isLinkedToController) {
			DHDTE cte = getLinkedControllerTE();
			if (cte != null)
				cte.clearLinkToStargate();
			clearLinkToController();
		}
	}
	
	public void clearLinkToController() {
		System.out.printf("SGBaseTE: Unlinking stargate at (%d, %d, %d) from controller\n",
			xCoord, yCoord, zCoord);
		isLinkedToController = false;
		markDirty();
	}

	//------------------------------------   Server   --------------------------------------------
	
	public void connectOrDisconnect(String address, EntityPlayer player) {
		if (debugConnect)
			System.out.printf("SGBaseTE: %s: connectOrDisconnect('%s') in state %s by %s\n",
				side(), address, state, player);
		if (address.length() > 0)
			connect(address, player);
		else {
			boolean canDisconnect = isInitiator || closeFromEitherEnd;
			SGBaseTE dte = getConnectedStargateTE();
			boolean validConnection =
				(dte != null) && (dte.getConnectedStargateTE() == this);
			if (canDisconnect || !validConnection) {
				if (state != SGState.Disconnecting)
					disconnect();
			}
			else
				if (!canDisconnect)
					System.out.printf("SGBaseTE.connectOrDisconnect: Not initiator\n");
		}
	}
	
	void connect(String address, EntityPlayer player) {
		String homeAddress = findHomeAddress();
		SGBaseTE dte;
		if (address.length() > getNumChevrons()) {
			diallingFailure(player, "Not enough chevrons to dial " + address);
			return;
		}
		try {
			dte = SGAddressing.findAddressedStargate(address, worldObj);
		}
		catch (SGAddressing.AddressingError e) {
			diallingFailure(player, e.getMessage());
			return;
		}
		if (dte == null) {
			diallingFailure(player, "No stargate at address " + address);
			return;
		}
		if (dte == this) {
			diallingFailure(player, "Stargate cannot connect to itself");
			return;
		}
		if (debugConnect)
			System.out.printf("SGBaseTE.connect: to (%s,%s,%s) in dimension %d with state %s\n",
				dte.xCoord, dte.yCoord, dte.zCoord, dte.getWorldObj().provider.dimensionId,
				dte.state);
		if (getWorldObj() == dte.getWorldObj()) {
			address = SGAddressing.localAddress(address);
			homeAddress = SGAddressing.localAddress(homeAddress);
		}
		if (dte.getNumChevrons() < homeAddress.length()) {
			diallingFailure(player, "Destination stargate has insufficient chevrons");
			return;
		}
		//System.out.printf("SGBaseTE.connect: addressed TE state = %s\n", dte.state);
		if (dte.state != SGState.Idle) {
			diallingFailure(player, "Stargate at address " + address + " is busy");
			return;
		}
		distanceFactor = distanceFactorForCoordDifference(this, dte);
		if (debugEnergyUse)
			System.out.printf("SGBaseTE: distanceFactor = %s\n", distanceFactor);
		if (!energyIsAvailable(energyToOpen * distanceFactor)) {
			diallingFailure(player, "Stargate has insufficient energy");
			return;
		}
		startDiallingStargate(address, dte, true);
		dte.startDiallingStargate(homeAddress, this, false);
	}
	
	double distanceFactorForCoordDifference(TileEntity te1, TileEntity te2) {
		double dx = te1.xCoord - te2.xCoord;
		double dz = te1.zCoord - te2.zCoord;
		double d = Math.sqrt(dx * dx + dz * dz);
		if (debugEnergyUse)
			System.out.printf("SGBaseTE: Connection distance = %s\n", d);
		double ld = Math.log(0.05 * d + 1);
		double lm = Math.log(0.05 * 16 * SGAddressing.coordRange);
		double lr = ld / lm;
		double f = 1 + 14 * distanceFactorMultiplier * lr * lr;
		if (te1.getWorldObj() != te2.getWorldObj())
			f *= interDimensionMultiplier;
		return f;
	}
	
	void diallingFailure(EntityPlayer player, String mess) {
		player.addChatMessage(new ChatComponentText(mess));
		playSoundEffect("gcewing_sg:sg_abort", 1.0F, 1.0F);
	}
	
	String findHomeAddress() {
		String homeAddress;
		try {
			return getHomeAddress();
		}
		catch (SGAddressing.AddressingError e) {
			System.out.printf("SGBaseTE.findHomeAddress: %s\n", e);
			return "";
		}
	}
	
	public void disconnect() {
		if (debugConnect)
			System.out.printf("SGBaseTE: %s: disconnect()\n", side());
		SGBaseTE dte = SGBaseTE.at(connectedLocation);
		if (dte != null)
			dte.clearConnection();
		clearConnection();
	}
	
	public void clearConnection() {
		if (state != SGState.Idle || connectedLocation != null) {
			dialledAddress = "";
			connectedLocation = null;
			isInitiator = false;
			firstEngagedChevron = 0;
			numEngagedChevrons = 0;
			markDirty();
			markBlockForUpdate();
			if (state == SGState.Connected) {
				enterState(SGState.Disconnecting, disconnectTime);
				//sendClientEvent(SGEvent.StartDisconnecting, 0);
				playSoundEffect("gcewing_sg:sg_close", 1.0F, 1.0F);
			}
			else {
				if (state != SGState.Idle && state != SGState.Disconnecting)
					playSoundEffect("gcewing_sg:sg_abort", 1.0F, 1.0F);
				enterState(SGState.Idle, 0);
				//sendClientEvent(SGEvent.FinishDisconnecting, 0);
			}
		}
	}
	
	void startDiallingStargate(String address, SGBaseTE dte, boolean initiator) {
		//System.out.printf("SGBaseTE.startDiallingStargate %s, initiator = %s\n",
		//	dte, initiator);
		dialledAddress = address;
		firstEngagedChevron = (getNumChevrons() - address.length()) / 2;
		connectedLocation = new SGLocation(dte);
		isInitiator = initiator;
		markDirty();
		startDiallingNextSymbol();
	}

	void serverUpdate() {
		if (isMerged) {
			if (debugState && state != SGState.Connected && timeout > 0) {
				int dimension = worldObj.provider.dimensionId;
				System.out.printf(
					"SGBaseTE.serverUpdate at (%d, %d, %d) in dimension %d: state %s, timeout %s\n",
					xCoord, yCoord, zCoord, dimension, state, timeout);
			}
			tickEnergyUsage();
			if (timeout > 0) {
				if (state == SGState.Transient)
					performTransientDamage();
				--timeout;
			}
			else switch(state) {
				case Idle:
					if (undialledDigitsRemaining())
						startDiallingNextSymbol();
					break;
				case Dialling:
					finishDiallingSymbol();
					break;
				case InterDialling:
					startDiallingNextSymbol();
					break;
				case Transient:
					enterState(SGState.Connected, isInitiator ? ticksToStayOpen : 0);
					break;
				case Connected:
					if (isInitiator)
						disconnect();
					break;
				case Disconnecting:
					enterState(SGState.Idle, 0);
					break;
			}
		}
	}
	
	void tickEnergyUsage() {
		if (state == SGState.Connected && isInitiator)
			if (!useEnergy(energyUsePerTick * distanceFactor))
				disconnect();
	}
	
	boolean energyIsAvailable(double amount) {
		List<ISGEnergySource> sources = findEnergySources();
		double energy = energyInBuffer + energyAvailableFrom(sources);
		if (debugEnergyUse)
			System.out.printf("SGBaseTE.energyIsAvailable: need %s, have %s\n", amount, energy);
		return energy >= amount;
	}
	
	boolean useEnergy(double amount) {
		if (debugEnergyUse)
			System.out.printf("SGBaseTE.useEnergy: %s; buffered: %s\n", amount, energyInBuffer);
		if (amount <= energyInBuffer) {
			energyInBuffer -= amount;
			return true;
		}
		List<ISGEnergySource> sources = findEnergySources();
		double energyAvailable = energyInBuffer + energyAvailableFrom(sources);
		if (debugEnergyUse)
			System.out.printf("SGBaseTE.useEnergy: %s available\n", energyAvailable);
		if (amount > energyAvailable) {
			System.out.printf("SGBaseTE: Not enough energy available\n");
			return false;
		}
		double desiredEnergy = max(amount, maxEnergyBuffer);
		double targetEnergy = min(desiredEnergy, energyAvailable);
		double energyRequired = targetEnergy - energyInBuffer;
		if (debugEnergyUse)
			System.out.printf("SGBaseTE.useEnergy: another %s required\n", energyRequired);
		double energyOnHand = energyInBuffer + drawEnergyFrom(sources, energyRequired);
		if (debugEnergyUse)
			System.out.printf("SGBaseTE.useEnergy: %s now on hand, need %s\n", energyOnHand, amount);
		if (amount - 0.0001 > energyOnHand) {
			System.out.printf("SGBaseTE: Energy sources only delivered %s of promised %s\n",
				energyOnHand - energyInBuffer, energyAvailable);
			return false;
		}
		setEnergyInBuffer(energyOnHand - amount);
		if (debugEnergyUse)
			System.out.printf("SGBaseTE.useEnergy: %s left over in buffer\n", energyInBuffer);
		return true;
	}
	
//	List<ISGEnergySource> findEnergySources() {
//		List<ISGEnergySource> result = new ArrayList<ISGEnergySource>();
//		DHDTE te = getLinkedControllerTE();
//		if (te != null)
//			result.add(te);
//		for (ForgeDirection d : ForgeDirection.VALID_DIRECTIONS) {
//			TileEntity nte = worldObj.getTileEntity(xCoord + d.offsetX, yCoord + d.offsetY, zCoord + d.offsetZ);
//			if (nte instanceof ISGEnergySource)
//				result.add((ISGEnergySource)nte);
//		}
//		return result;
//	}
	
	List<ISGEnergySource> findEnergySources() {
		//System.out.printf("SGBaseTe.findEnergySources: for (%d,%d,%d)\n",
		//	xCoord, yCoord, zCoord);
		List<ISGEnergySource> result = new ArrayList<ISGEnergySource>();
		DHDTE te = getLinkedControllerTE();
		if (te != null)
			result.add(te);
		int dz = getBlockMetadata() & 1;
		int dx = 1 - dz;
		for (int i = -2; i <= 2; i++) {
			System.out.printf("SGBaseTe.findEnergySources: Checking (%d,%d,%d)\n",
				xCoord + i * dx, yCoord - 1, zCoord + i * dz);
			TileEntity nte = worldObj.getTileEntity(xCoord + i * dx, yCoord - 1, zCoord + i * dz);
			if (nte instanceof ISGEnergySource)
				result.add((ISGEnergySource)nte);
		}
		return result;
	}
	
	double energyAvailableFrom(List<ISGEnergySource> sources) {
		double energy = 0;
		for (ISGEnergySource source : sources) {
			double e = source.availableEnergy();
			if (debugEnergyUse)
				System.out.printf("SGBaseTe.energyAvailableFrom: %s can supply %s\n", source, e);
			energy += e;
		}
		return energy;
	}
	
	double drawEnergyFrom(List<ISGEnergySource> sources, double amount) {
		double total = 0;
		for (ISGEnergySource source : sources) {
			if (total >= amount)
				break;
			double e = source.drawEnergy(amount - total);
			if (debugEnergyUse)
				System.out.printf("SGBaseTe.drawEnergyFrom: %s supplied %s\n", source, e);
			total += e;
		}
		if (total < amount)
			System.out.printf("SGCraft: Warning: Energy sources did not deliver promised energy " +
				"(%s requested, %s delivered)\n", amount, total);
		return total;
	}
	
	void setEnergyInBuffer(double amount) {
		if (energyInBuffer != amount) {
			energyInBuffer = amount;
			markDirty();
		}
	}
	
	public Trans3 localToGlobalTransformation() {
		return getBlock().localToGlobalTransformation(xCoord, yCoord, zCoord, getBlockMetadata(), this);
	}
	
	void performTransientDamage() {
		Trans3 t = localToGlobalTransformation();
		Vector3 p0 = t.p(-1.5, 0.5, 0.5);
		Vector3 p1 = t.p(1.5, 3.5, 5.5);
		Vector3 q0 = p0.min(p1);
		Vector3 q1 = p0.max(p1);
		AxisAlignedBB box = AxisAlignedBB.getBoundingBox(q0.x, q0.y, q0.z, q1.x, q1.y, q1.z);
		//System.out.printf("SGBaseTE.performTransientDamage: players in world:\n");
		//for (Entity ent : (List<Entity>)worldObj.loadedEntityList)
		//	if (ent instanceof EntityPlayer)
		//		System.out.printf("--- %s\n", ent);
		//System.out.printf("SGBaseTE.performTransientDamage: box = %s\n", box);
		List<EntityLiving> ents = worldObj.getEntitiesWithinAABB(EntityLiving.class, box);
		//System.out.printf("SGBaseTE.performTransientDamage: entities in box:\n", box);
		for (EntityLiving ent : ents) {
			Vector3 ep = new Vector3(ent.posX, ent.posY, ent.posZ);
			Vector3 gp = t.p(0, 2, 0.5);
			double dist = ep.distance(gp);
			System.out.printf("SGBaseTE.performTransientDamage: found %s\n", ent);
			if (dist > 1.0)
				dist = 1.0;
			int damage = (int)Math.ceil(dist * transientDamageRate);
			System.out.printf("SGBaseTE.performTransientDamage: distance = %s, damage = %s\n",
				dist, damage);
			ent.attackEntityFrom(transientDamage, damage);
		}
	}
	
	boolean undialledDigitsRemaining() {
		int n = numEngagedChevrons;
		return n < dialledAddress.length();
	}
	
	void startDiallingNextSymbol() {
		if (debugState)
			System.out.printf("SGBaseTE.startDiallingNextSymbol: %s of %s\n",
				numEngagedChevrons, dialledAddress);
		startDiallingSymbol(dialledAddress.charAt(numEngagedChevrons));
	}
	
	void startDiallingSymbol(char c) {
		int i = SGAddressing.charToSymbol(c);
		if (debugState)
			System.out.printf("SGBaseTE.startDiallingSymbol: %s\n", i);
		if (i >= 0 && i < numRingSymbols) {
			int chevronNo = firstEngagedChevron + numEngagedChevrons;
			startDiallingToAngle(i * ringSymbolAngle - 45 * chevronNo);
			playSoundEffect(diallingSound[diallingSpeed()], 1.0F, 1.0F);
		}
		else {
			System.out.printf("SGCraft: Stargate jammed trying to dial symbol %s\n", c);
			dialledAddress = "";
			enterState(SGState.Idle, 0);
		}
	}
	
	void startDiallingToAngle(double a) {
		targetRingAngle = Utils.normaliseAngle(a);
		enterState(SGState.Dialling, diallingTime[diallingSpeed()]);
	}
	
	void finishDiallingSymbol() {
		++numEngagedChevrons;
		if (undialledDigitsRemaining())
			enterState(SGState.InterDialling, interDiallingTime[diallingSpeed()]);
		else
			finishDiallingAddress();
	}
	
	int diallingSpeed() {
//		if (dialledAddress.length() == SGAddressing.maxAddressLength)
//			return 1;
//		else
			return 0;
	}
	
	void finishDiallingAddress() {
		//System.out.printf("SGBaseTE: Connecting to '%s'\n", dialledAddress);
		if (!isInitiator || useEnergy(energyToOpen * distanceFactor)) {
			enterState(SGState.Transient, transientDuration);
			playSoundEffect("gcewing_sg:sg_open", 1.0F, 1.0F);
		}
		else
			disconnect();
	}
	
	boolean canTravelFromThisEnd() {
		return isInitiator || !oneWayTravel;
	}
	
	static String repr(Entity entity) {
		if (entity != null) {
			String s = String.format("%s#%s", entity.getClass().getSimpleName(), entity.getEntityId());
			if (entity.isDead)
				s += "(dead)";
			return s;
		}
		else
			return "null";
	}
	
	class TrackedEntity {
		public Entity entity;
		public Vector3 lastPos;
		
		public TrackedEntity(Entity entity) {
			this.entity = entity;
			this.lastPos = new Vector3(entity.posX, entity.posY, entity.posZ);
		}
		
	}

	List<TrackedEntity> trackedEntities = new ArrayList<TrackedEntity>();
	
	void checkForEntitiesInPortal() {
		if (state == SGState.Connected) {
			for (TrackedEntity trk : trackedEntities)
				entityInPortal(trk.entity, trk.lastPos);
			trackedEntities.clear();
			Vector3 p0 = new Vector3(-1.5, 0.5, -3.5);
			Vector3 p1 = new Vector3(1.5, 3.5, 3.5);
			Trans3 t = localToGlobalTransformation();
			AxisAlignedBB box = t.box(p0, p1);
			//System.out.printf("SGBaseTE.checkForEntitiesInPortal: %s\n", box);
			List<Entity> ents = (List<Entity>)worldObj.getEntitiesWithinAABB(Entity.class, box);
			for (Entity entity : ents) {
				if (!entity.isDead && entity.ridingEntity == null) {
					//if (!(entity instanceof EntityPlayer))
					//	System.out.printf("SGBaseTE.checkForEntitiesInPortal: Tracking %s\n", repr(entity));
					trackedEntities.add(new TrackedEntity(entity));
				}
			}
		}
	}

	public void entityInPortal(Entity entity, Vector3 prevPos) {
		if (!entity.isDead && state == SGState.Connected && canTravelFromThisEnd()) {
			Trans3 t = localToGlobalTransformation();
			double vx = entity.posX - prevPos.x;
			double vy = entity.posY - prevPos.y;
			double vz = entity.posZ - prevPos.z;
			Vector3 p1 = t.ip(entity.posX, entity.posY, entity.posZ);
			Vector3 p0 = t.ip(2 * prevPos.x - entity.posX, 2 * prevPos.y - entity.posY, 2 * prevPos.z - entity.posZ);
			//if (!(entity instanceof EntityPlayer))
			//	System.out.printf("SGBaseTE.entityInPortal: z0 = %.3f z1 = %.3f\n", p0.z, p1.z);
			double z0 = 0.0;
			if (p0.z >= z0 && p1.z < z0) {
				//System.out.printf("SGBaseTE.entityInPortal: %s passed through event horizon of stargate at (%d,%d,%d) in %s\n",
				//	repr(entity), xCoord, yCoord, zCoord, worldObj);
				entity.motionX = vx;
				entity.motionY = vy;
				entity.motionZ = vz;
				//System.out.printf("SGBaseTE.entityInPortal: %s pos (%.2f, %.2f, %.2f) prev (%.2f, %.2f, %.2f) motion (%.2f, %.2f, %.2f)\n",
				//	repr(entity),
				//	entity.posX, entity.posY, entity.posZ,
				//	prevPos.x, prevPos.y, prevPos.z,
				//	entity.motionX, entity.motionY, entity.motionZ);
				SGBaseTE dte = getConnectedStargateTE();
				if (dte != null) {
					Trans3 dt = dte.localToGlobalTransformation();
					while (entity.ridingEntity != null)
						entity = entity.ridingEntity;
					teleportEntityAndRider(entity, t, dt, connectedLocation.dimension);
				}
			}
		}
	}
	
	Entity teleportEntityAndRider(Entity entity, Trans3 t1, Trans3 t2, int dimension) {
		Entity rider = entity.riddenByEntity;
		if (rider != null) {
			//System.out.printf("SGBaseTE.teleportEntityAndRider: Unmounting %s from %s\n",
			//	repr(rider), repr(entity));
			rider.mountEntity(null);
			rider = teleportEntityAndRider(rider, t1, t2, dimension);
		}
		entity = teleportEntity(entity, t1, t2, dimension);
		if (rider != null) {
			//System.out.printf("SGBaseTE.teleportEntityAndRider: Mounting %s on %s\n",
			//	repr(rider), repr(entity));
			rider.mountEntity(entity);
		}
		return entity;
	}
	
	static Entity teleportEntity(Entity entity, Trans3 t1, Trans3 t2, int dimension) {
		Entity newEntity;
		//System.out.printf("SGBaseTE.teleportEntity: %s (in dimension %d)  to dimension %d\n",
		//	repr(entity), entity.dimension, dimension);
		//System.out.printf("SGBaseTE.teleportEntity: pos (%.2f, %.2f, %.2f) prev (%.2f, %.2f, %.2f) last (%.2f, %.2f, %.2f)\n",
		//	entity.posX, entity.posY, entity.posZ,
		//	entity.prevPosX, entity.prevPosY, entity.prevPosZ,
		//	entity.lastTickPosX, entity.lastTickPosY, entity.lastTickPosZ);
		Vector3 p = t1.ip(entity.posX, entity.posY, entity.posZ); // local position
		Vector3 v = t1.iv(entity.motionX, entity.motionY, entity.motionZ); // local velocity
		Vector3 r = t1.iv(yawVector(entity)); // local facing
		Vector3 q = t2.p(-p.x, p.y, -p.z); // new global position
		Vector3 u = t2.v(-v.x, v.y, -v.z); // new global velocity
		Vector3 s = t2.v(r.mul(-1)); // new global facing
		double a = yawAngle(s); // new global yaw angle
		if (entity.dimension == dimension)
			newEntity = teleportWithinDimension(entity, q, u, a);
		else {
			newEntity = teleportToOtherDimension(entity, q, u, a, dimension);
			if (newEntity != null)
				newEntity.dimension = dimension;
			else
				System.out.printf("SGBaseTE.teleportEntity: teleportToOtherDimension returned null for %s\n",
					entity);
		}
		//if (entity != newEntity)
		//	System.out.printf("SGBaseTE.teleportEntity: %s is now %s\n", repr(entity), repr(newEntity));
		return newEntity;
	}

	static Entity teleportWithinDimension(Entity entity, Vector3 p, Vector3 v, double a) {
		if (entity instanceof EntityPlayerMP)
			return teleportPlayerWithinDimension((EntityPlayerMP)entity, p, v, a);
		else
			return teleportEntityToWorld(entity, p, v, a, (WorldServer)entity.worldObj);
	}
	
	static Entity teleportPlayerWithinDimension(EntityPlayerMP entity, Vector3 p, Vector3 v, double a) {
		entity.rotationYaw = (float)a;
		entity.setPositionAndUpdate(p.x, p.y, p.z);
		entity.worldObj.updateEntityWithOptionalForce(entity, false);
		return entity;
	}

	static Entity teleportToOtherDimension(Entity entity, Vector3 p, Vector3 v, double a, int dimension) {
		if (entity instanceof EntityPlayerMP) {
			EntityPlayerMP player = (EntityPlayerMP)entity;
			Vector3 q = p.add(yawVector(a));
			transferPlayerToDimension(player, dimension, q, a);
			return player;
		}
		else
			return teleportEntityToDimension(entity, p, v, a, dimension);
	}
	
	static void sendDimensionRegister(EntityPlayerMP player, int dimensionID) {
		int providerID = DimensionManager.getProviderType(dimensionID);
		ForgeMessage msg = new ForgeMessage.DimensionRegisterMessage(dimensionID, providerID);
		FMLEmbeddedChannel channel = NetworkRegistry.INSTANCE.getChannel("FORGE", Side.SERVER);
		channel.attr(FMLOutboundHandler.FML_MESSAGETARGET).set(FMLOutboundHandler.OutboundTarget.PLAYER);
		channel.attr(FMLOutboundHandler.FML_MESSAGETARGETARGS).set(player);
		channel.writeAndFlush(msg).addListener(ChannelFutureListener.FIRE_EXCEPTION_ON_FAILURE);
	}
	
	static void transferPlayerToDimension(EntityPlayerMP player, int newDimension, Vector3 p, double a) {
		//System.out.printf("SGBaseTE.transferPlayerToDimension: %s to dimension %d\n", repr(player), newDimension);
		MinecraftServer server = MinecraftServer.getServer();
		ServerConfigurationManager scm = server.getConfigurationManager();
		int oldDimension = player.dimension;
		player.dimension = newDimension;
		WorldServer oldWorld = server.worldServerForDimension(oldDimension);
		WorldServer newWorld = server.worldServerForDimension(newDimension);
		//System.out.printf("SGBaseTE.transferPlayerToDimension: %s with %s\n", newWorld, newWorld.getEntityTracker());
		// <<< Fix for MCPC+
		//Packet regdim = ForgePacket.makePacketSet(new DimensionRegisterPacket(newDimension,
		//	DimensionManager.getProviderType(newDimension)))[0];
		//player.playerNetServerHandler.sendPacketToPlayer(regdim);
		sendDimensionRegister(player, newDimension);
		// >>>
		player.closeScreen();
		player.playerNetServerHandler.sendPacket(new S07PacketRespawn(player.dimension,
			player.worldObj.difficultySetting, newWorld.getWorldInfo().getTerrainType(),
			player.theItemInWorldManager.getGameType()));
		oldWorld.removePlayerEntityDangerously(player); // Removes player right now instead of waiting for next tick
		player.isDead = false;
		//player.setLocationAndAngles(player.posX, player.posY, player.posZ, player.rotationYaw, player.rotationPitch);
		player.setLocationAndAngles(p.x, p.y, p.z, (float)a, player.rotationPitch);
		newWorld.spawnEntityInWorld(player);
		player.setWorld(newWorld);
		//newWorld.updateEntityWithOptionalForce(player, false);
		scm.func_72375_a(player, oldWorld);
		player.playerNetServerHandler.setPlayerLocation(p.x, p.y, p.z, (float)a, player.rotationPitch);
		player.theItemInWorldManager.setWorld(newWorld);
		scm.updateTimeAndWeatherForPlayer(player, newWorld);
		scm.syncPlayerInventory(player);
		Iterator var6 = player.getActivePotionEffects().iterator();
		while (var6.hasNext()) {
			PotionEffect effect = (PotionEffect)var6.next();
			player.playerNetServerHandler.sendPacket(new S1DPacketEntityEffect(player.getEntityId(), effect));
		}
		player.playerNetServerHandler.sendPacket(new S1FPacketSetExperience(player.experience, player.experienceTotal, player.experienceLevel));
		//GameRegistry.onPlayerChangedDimension(player); // Obsolete?
		//System.out.printf("SGBaseTE.transferPlayerToDimension: Transferred %s\n", repr(player));
	}	
	
	static Entity teleportEntityToDimension(Entity entity, Vector3 p, Vector3 v, double a, int dimension) {
		//System.out.printf("SGBaseTE.teleportEntityToDimension: %s to dimension %d\n", repr(entity), dimension);
		MinecraftServer server = MinecraftServer.getServer();
		WorldServer world = server.worldServerForDimension(dimension);
		return teleportEntityToWorld(entity, p, v, a, world);
	}
	
	static Entity teleportEntityToWorld(Entity oldEntity, Vector3 p, Vector3 v, double a, WorldServer newWorld) {
		//System.out.printf("SGBaseTE.teleportEntityToWorld: %s to %s\n", repr(oldEntity), newWorld);
		WorldServer oldWorld = (WorldServer)oldEntity.worldObj;
		NBTTagCompound nbt = new NBTTagCompound();
		//oldEntity.writeToNBTOptional(nbt);
		oldEntity.writeToNBT(nbt);
		extractEntityFromWorld(oldWorld, oldEntity);
		//Entity newEntity = EntityList.createEntityFromNBT(nbt, newWorld);
		Entity newEntity = instantiateEntityFromNBT(oldEntity.getClass(), nbt, newWorld);
		if (newEntity != null) {
			if (oldEntity instanceof EntityLiving)
				copyMoreEntityData((EntityLiving)oldEntity, (EntityLiving)newEntity);
			setVelocity(newEntity, v);
			//System.out.printf("SGBaseTE.teleportEntityToWorld: Set velocity of %s to (%.2f, %.2f, %.2f)\n",
			//	repr(newEntity), newEntity.motionX, newEntity.motionY, newEntity.motionZ);
			newEntity.setLocationAndAngles(p.x, p.y, p.z, (float)a, oldEntity.rotationPitch);
			checkChunk(newWorld, newEntity);
			//System.out.printf("SGBaseTE.teleportEntityToWorld: Spawning %s in %s\n", repr(newEntity), newWorld);
			newEntity.forceSpawn = true; // Force spawn packet to be sent as soon as possible
			newWorld.spawnEntityInWorld(newEntity);
			newEntity.setWorld(newWorld);
			//System.out.printf(
			//	"SGBaseTE.teleportEntityToWorld: Spawned %s pos (%.2f, %.2f, %.2f) vel (%.2f, %.2f, %.2f)\n",
			//	repr(newEntity),
			//	newEntity.posX, newEntity.posY, newEntity.posZ,
			//	newEntity.motionX, newEntity.motionY, newEntity.motionZ);
		}
		oldWorld.resetUpdateEntityTick();
		if (oldWorld != newWorld)
			newWorld.resetUpdateEntityTick();
		return newEntity;
	}
	
	static Entity instantiateEntityFromNBT(Class cls, NBTTagCompound nbt, WorldServer world) {
		try {
			Entity entity = (Entity)cls.getConstructor(World.class).newInstance(world);
			entity.readFromNBT(nbt);
			return entity;
		}
		catch (Exception e) {
			System.out.printf("SGCraft: SGBaseTE.instantiateEntityFromNBT: Could not instantiate %s: %s\n",
				cls, e);
			e.printStackTrace();
			return null;
		}
	}

	static void copyMoreEntityData(EntityLiving oldEntity, EntityLiving newEntity) {
		float s = oldEntity.getAIMoveSpeed();
		if (s != 0)
			newEntity.setAIMoveSpeed(s);
	}

	static void setVelocity(Entity entity, Vector3 v) {
		entity.motionX = v.x;
		entity.motionY = v.y;
		entity.motionZ = v.z;
	}
	
	static void extractEntityFromWorld(World world, Entity entity) {
		// Immediately remove entity from world without calling setDead(), which has
		// undesirable side effects on some entities.
		if (entity instanceof EntityPlayer) {
			world.playerEntities.remove(entity);
			world.updateAllPlayersSleepingFlag();
		}
		int i = entity.chunkCoordX;
		int j = entity.chunkCoordZ;
		if (entity.addedToChunk && world.getChunkProvider().chunkExists(i, j))
			world.getChunkFromChunkCoords(i, j).removeEntity(entity);
		world.loadedEntityList.remove(entity);
		//SGWorldAccess.onEntityRemoved(world, entity);
		BaseReflectionUtils.call(world, onEntityRemoved, entity);
	}
	
	static void checkChunk(World world, Entity entity) {
		int cx = MathHelper.floor_double(entity.posX / 16.0D);
		int cy = MathHelper.floor_double(entity.posZ / 16.0D);
		Chunk chunk = world.getChunkFromChunkCoords(cx, cy);
	}
	
	static Vector3 yawVector(Entity entity) {
		return yawVector(entity.rotationYaw);
	}
	
	static Vector3 yawVector(double yaw) {
		double a = Math.toRadians(yaw);
		Vector3 v = new Vector3(-Math.sin(a), 0, Math.cos(a));
		//System.out.printf("SGBaseTE.yawVector: %.2f --> (%.3f, %.3f)\n", yaw, v.x, v.z);
		return v;
	}
	
	static double yawAngle(Vector3 v) {
		double a = Math.atan2(-v.x, v.z);
		double d = Math.toDegrees(a);
		//System.out.printf("SGBaseTE.yawAngle: (%.3f, %.3f) --> %.2f\n", v.x, v.z, d);
		return d;
	}
	
	SGBaseTE getConnectedStargateTE() {
		if (connectedLocation != null)
			return connectedLocation.getStargateTE();
		else
			return null;
		}
	
	//------------------------------------   Client   --------------------------------------------

	@Override
	public void onDataPacket(NetworkManager net, S35PacketUpdateTileEntity pkt) {
		//System.out.printf("SGBaseTE.onDataPacket: with state %s numEngagedChevrons %s\n",
		//	SGState.valueOf(pkt.customParam1.getInteger("state")),
		//	pkt.customParam1.getInteger("numEngagedChevrons"));
		SGState oldState = state;
		super.onDataPacket(net, pkt);
		if (isMerged && state != oldState) {
			switch (state) {
				case Transient:
					initiateOpeningTransient();
					break;
				case Disconnecting:
					initiateClosingTransient();
					break;
			}
		}
	}
	
	void clientUpdate() {
		lastRingAngle = ringAngle;
		switch (state) {
			case Dialling:
				//System.out.printf("SGBaseTe: Relaxing angle %s towards %s at rate %s\n",
				//	ringAngle, targetRingAngle, diallingRelaxationRate);
				//setRingAngle(Utils.relaxAngle(ringAngle, targetRingAngle, diallingRelaxationRate));
				updateRingAngle();
				//System.out.printf("SGBaseTe: Ring angle now %s\n", ringAngle);
				break;
			case Transient:
			case Connected:
			case Disconnecting:
				applyRandomImpulse();
				updateEventHorizon();
				break;
		}
	}

	void setRingAngle(double a) {
		ringAngle = a;
	}

	void updateRingAngle() {
		if (timeout > 0) {
			double da = Utils.diffAngle(ringAngle, targetRingAngle) / timeout;
			setRingAngle(Utils.addAngle(ringAngle, da));
			--timeout;
		}
		else
			setRingAngle(targetRingAngle);
	}
	
	public double[][][] getEventHorizonGrid() {
		if (ehGrid == null) {
			int m = SGBaseTERenderer.ehGridRadialSize;
			int n = SGBaseTERenderer.ehGridPolarSize;
			ehGrid = new double[2][n + 2][m + 1];
			for (int i = 0; i < 2; i++) {
				ehGrid[i][0] = ehGrid[i][n];
				ehGrid[i][n + 1] = ehGrid[i][1];
			}
		}
		return ehGrid;
	}
	
	void initiateOpeningTransient() {
		double v[][] = getEventHorizonGrid()[1];
		int n = SGBaseTERenderer.ehGridPolarSize;
		for (int j = 0; j <= n+1; j++) {
			v[j][0] = openingTransientIntensity;
			v[j][1] = v[j][0] + openingTransientRandomness * random.nextGaussian();
		}
	}
	
	void initiateClosingTransient() {
		double v[][] = getEventHorizonGrid()[1];
		int m = SGBaseTERenderer.ehGridRadialSize;
		int n = SGBaseTERenderer.ehGridPolarSize;
		for (int i = 1; i < m; i++)
			for (int j = 1; j <= n; j++)
				v[j][i] += closingTransientRandomness * random.nextGaussian();
	}
	
	void applyRandomImpulse() {
		double v[][] = getEventHorizonGrid()[1];
		int m = SGBaseTERenderer.ehGridRadialSize;
		int n = SGBaseTERenderer.ehGridPolarSize;
		int i = random.nextInt(m - 1) + 1;
		int j = random.nextInt(n) + 1;
		v[j][i] += 0.05 * random.nextGaussian();
	}
	
	void updateEventHorizon() {
		double grid[][][] = getEventHorizonGrid();
		double u[][] = grid[0];
		double v[][] = grid[1];
		int m = SGBaseTERenderer.ehGridRadialSize;
		int n = SGBaseTERenderer.ehGridPolarSize;
		double dt = 1.0;
		double asq = 0.03;
		double d = 0.95;
		for (int i = 1; i < m; i++)
			for (int j = 1; j <= n; j++) {
				double du_dr = 0.5 * (u[j][i+1] - u[j][i-1]);
				double d2u_drsq = u[j][i+1] - 2 * u[j][i] + u[j][i-1];
				double d2u_dthsq = u[j+1][i] - 2 * u[j][i] + u[j-1][i];
				v[j][i] = d * v[j][i] + (asq * dt) * (d2u_drsq + du_dr / i + d2u_dthsq / (i * i));
		}
		for (int i = 1; i < m; i++)
			for (int j = 1; j <= n; j++)
				u[j][i] += v[j][i] * dt;
		double u0 = 0, v0 = 0;
		for (int j = 1; j <= n; j++) {
			u0 += u[j][1];
			v0 += v[j][1];
		}
		u0 /= n;
		v0 /= n;
		for (int j = 1; j <= n; j++) {
			u[j][0] = u0;
			v[j][0] = v0;
		}
		//dumpGrid("u", u);
		//dumpGrid("v", v);
	}
	
	void dumpGrid(String label, double g[][]) {
		System.out.printf("SGBaseTE: %s:\n", label);
		int m = SGBaseTERenderer.ehGridRadialSize;
		int n = SGBaseTERenderer.ehGridPolarSize;
		for (int j = 0; j <= n+1; j++) {
			for (int i = 0; i <= m; i++)
				System.out.printf(" %6.3f", g[j][i]);
			System.out.printf("\n");
		}
	}
	
	@Override
	BaseTEChunkManager getChunkManager() {
		return SGCraft.chunkManager;
	}

//	@Override
//	IInventory getInventory() {
//		return inventory;
//	}
	
}

//------------------------------------------------------------------------------------------------

class TransientDamageSource extends DamageSource {

	public TransientDamageSource() {
		super("sgTransient");
	}
	
	public String getDeathMessage(EntityPlayer player) {
		return player.getCommandSenderName() + " was torn apart by an event horizon";
	}
	
}

//------------------------------------------------------------------------------------------------

//class PendingRemount {
//	public Entity rider;
//	public Entity ridden;
//	public int delay;
//	
//	public PendingRemount(Entity rider, Entity ridden) {
//		this.rider = rider;
//		this.ridden = ridden;
//		this.delay = 4;
//	}
//
//}
