//------------------------------------------------------------------------------------------------
//
//   SG Craft - IC2 Stargate Power Unit Tile Entity
//
//------------------------------------------------------------------------------------------------

package gcewing.sg.ic2;

import net.minecraft.nbt.*;
import net.minecraft.tileentity.*;
import net.minecraftforge.common.*;
import net.minecraftforge.common.util.*;

import ic2.api.energy.event.*;
import ic2.api.energy.tile.*;

import gcewing.sg.*;
import static gcewing.sg.BaseUtils.*;
import static gcewing.sg.Utils.*;

public class IC2PowerTE extends BaseTileEntity implements IEnergySink, ISGEnergySource {

	boolean debugLoad = false;
	boolean debugInput = false;
	boolean debugOutput = false;

	final static int maxSafeInput = 2048;
	final static int maxEnergyBuffer = 1000000;
	final static double euPerSGEnergyUnit = 20.0;

	boolean loaded = false;
	double energyBuffer = 0;
	
	@Override
	public void updateEntity() {
		load();
	}
	
	@Override
	public void invalidate() {
		unload();
		super.invalidate();
	}
	
	@Override
	public void onChunkUnload() {
		unload();
		super.onChunkUnload();
	}
	
	void load() {
		if (!worldObj.isRemote && !loaded) {
			if(debugLoad)
				System.out.printf("SGCraft: IC2PowerTE: Adding to energy network\n");
			loaded = true;
			MinecraftForge.EVENT_BUS.post(new EnergyTileLoadEvent(this));
		}
	}			
	
	void unload() {
		if (!worldObj.isRemote && loaded) {
			if(debugLoad)
				System.out.printf("SGCraft: IC2PowerTE: Removing from energy network\n");
			MinecraftForge.EVENT_BUS.post(new EnergyTileUnloadEvent(this));
			loaded = false;
		}
	}
	
	@Override
	public void readContentsFromNBT(NBTTagCompound nbt) {
		super.readContentsFromNBT(nbt);
		energyBuffer = nbt.getDouble("energyBuffer");
	}
	
	public void writeContentsToNBT(NBTTagCompound nbt) {
		super.writeContentsToNBT(nbt);
		nbt.setDouble("energyBuffer", energyBuffer);
	}
	
	//------------------------- IEnergySink -------------------------
	
	@Override
	public boolean acceptsEnergyFrom(TileEntity emitter, ForgeDirection direction) {
		return true;
	}
	
	@Override
	public double demandedEnergyUnits() {
		double eu = min(maxEnergyBuffer - energyBuffer, maxSafeInput);
		if(debugInput)
			System.out.printf("SGCraft: IC2PowerTE: Demanding %s EU\n", eu);
		return eu;
	}
	
	@Override
	public double injectEnergyUnits(ForgeDirection directionFrom, double amount) {
		energyBuffer += amount;
		markDirty();
		markBlockForUpdate();
		if(debugInput)
			System.out.printf("SGCraft: IC2PowerTE: Injected %s EU giving %s\n", amount, energyBuffer);
		return 0;
	}
	
	@Override
	public int getMaxSafeInput() {
		return maxSafeInput;
	}
	
	//------------------------- ISGEnergySource -------------------------
	
	@Override
	public double availableEnergy() {
		double available = energyBuffer / euPerSGEnergyUnit;
		if (debugOutput)
			System.out.printf("SGCraft: IC2PowerTE: %s SGU available\n", available);
		return available;
	}
	
	public double drawEnergy(double request) {
		double available = energyBuffer / euPerSGEnergyUnit;
		double supply = min(request, available);
		energyBuffer -= supply * euPerSGEnergyUnit;
		markDirty();
		markBlockForUpdate();
		if(debugOutput)
			System.out.printf("SGCraft: IC2PowerTE: Supplying %s SGU of %s requested\n", supply, request);
		return supply;
	}

}
